#!/usr/bin/python
"""Wrapper script to run FAMLI on one or more FASTQ files."""

import os
import sys
import uuid
import time
import gzip
import json
import shutil
import logging
import argparse
from famli.exec_helpers import run_cmds
from famli.exec_helpers import align_reads
from famli.exec_helpers import return_results
from famli.exec_helpers import exit_and_clean_up
from famli.exec_helpers import get_reference_database
from famli.fastq_helpers import get_reads_from_url
from famli.fastq_helpers import set_up_sra_cache_folder
from famli.fastq_helpers import count_fastq_reads
from famli.fastq_helpers import combine_fastqs
from famli.famli_helpers import parse_alignment


class FAMLI:
    def __init__(self):
        parser = argparse.ArgumentParser(description="""
        Two options for running FAMLI:
        (1) Align a set of WGS reads - align against a reference database
        with DIAMOND, process those alignments with FAMLI, and save the
        results; or (2) Filter a set of existing alignments with FAMLI and
        save the results.""")

        parser.add_argument("command", help="""Command to run: align or filter. Invoke
            famli.py <command> -h for further details""")
        if len(sys.argv) < 2:
            parser.print_help()
        else:
            args = parser.parse_args(sys.argv[1:2])

            # Run the command that was specified
            if args.command == "align":
                self.align()
            elif args.command == "filter":
                self.filter()
            else:
                parser.print_help()
                print("Unrecognized command")

    def align(self):
        """Align a set of reads with DIAMOND and run FAMLI."""
        parser = argparse.ArgumentParser(
                description="""Align a set of reads with DIAMOND, filter alignments with FAMLI,
                and return the results""")

        parser.add_argument("--input",
                            type=str,
                            required=True,
                            help="""Location for input file(s). Combine multiple files with +.
                                    (Supported: sra://, s3://, or ftp://).""")
        parser.add_argument("--sample-name",
                            type=str,
                            required=True,
                            help="""Name of sample, sets output filename.""")
        parser.add_argument("--ref-db",
                            type=str,
                            required=True,
                            help="""Folder containing reference database.
                                    (Supported: s3://, ftp://, or local path).
                                    """)
        parser.add_argument("--output-folder",
                            type=str,
                            required=True,
                            help="""Folder to place results.
                                    (Supported: s3://, or local path).""")
        parser.add_argument("--min-score",
                            type=float,
                            default=20,
                            help="Minimum alignment score to report.")
        parser.add_argument("--blocks",
                            type=int,
                            default=5,
                            help="""Number of blocks used when aligning.
                                  Value relates to the amount of memory used.
                                  Roughly 6Gb RAM used by DIAMOND per block.
                                  """)
        parser.add_argument("--query-gencode",
                            type=int,
                            default=11,
                            help="Genetic code used to translate nucleotides.")
        parser.add_argument("--threads",
                            type=int,
                            default=16,
                            help="Number of threads to use aligning.")
        parser.add_argument("--min-qual",
                            type=int,
                            default=None,
                            help="Trim reads to a minimum Q score.")
        parser.add_argument("--temp-folder",
                            type=str,
                            default='/share',
                            help="Folder used for temporary files.")
        parser.add_argument("--batchsize",
                            type=int,
                            help="""Number of reads to process at a time.""")

        args = parser.parse_args(sys.argv[2:])

        # Make sure that there are no commas or whitespaces in the input
        input_str = args.input
        assert ' ' not in input_str, input_str
        assert ',' not in input_str, input_str

        # Make a temporary folder for all files to be placed in
        temp_folder = os.path.join(args.temp_folder, str(uuid.uuid4())[:8])
        assert os.path.exists(temp_folder) is False
        os.mkdir(temp_folder)

        # Set up logging
        log_fp = os.path.join(temp_folder, "log.txt")
        logFormatter = logging.Formatter(
            '%(asctime)s %(levelname)-8s [FAMLI] %(message)s'
        )
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.INFO)

        # Write to file
        fileHandler = logging.FileHandler(log_fp)
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)
        # Also write to STDOUT
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(logFormatter)
        rootLogger.addHandler(consoleHandler)

        # Check to see if DIAMOND is available
        logging.info("Checking for a working copy of DIAMOND")
        run_cmds(["diamond", "--version"])

        # Get the reference database
        try:
            db_fp = get_reference_database(
                args.ref_db,
                temp_folder
            )
        except:
            exit_and_clean_up(temp_folder)

        # Set up the NCBI fastq-dump cache folder within the temp folder
        set_up_sra_cache_folder(temp_folder)

        logging.info("Reference database: " + db_fp)

        # Align the input data and calculate the overall abundance

        # Keep track of the time elapsed to process this sample
        start_time = time.time()

        logging.info("Processing input argument: " + input_str)

        # Multiple input reads may be separated with a '+'
        input_str = input_str.split("+")
        # Make sure that they are all unique arguments
        assert len(input_str) == len(set(input_str)), "Duplicate arguments"
        # Make sure that the filenames are also all unique
        assert len(input_str) == len(set([
            s.split('/')[-1] for s in input_str
        ])), "Duplicate filenames"

        # Capture each command in a try statement
        # Get the input reads
        read_fps = []
        for s in input_str:
            logging.info("Fetching {}".format(s))
            try:
                read_fps.append(get_reads_from_url(
                    s, temp_folder, min_qual=args.min_qual))
            except:
                exit_and_clean_up(temp_folder)

        # Combine the files into a single FASTQ
        read_fp = os.path.join(temp_folder, "input.fastq")
        combine_fastqs(read_fps, read_fp)

        # Run the alignment
        try:
            align_fp = align_reads(
                read_fp,               # FASTQ file path
                db_fp,                 # Local path to DB
                temp_folder,    # Folder for results
                query_gencode=args.query_gencode,
                threads=args.threads,
                min_score=args.min_score,
                blocks=args.blocks,
            )
        except:
            exit_and_clean_up(temp_folder)

        # Process the alignments, reassigning multi-mapped reads
        try:
            with open(align_fp, "rt") as align_handle:
                aligned_reads, abund = parse_alignment(
                    align_handle,
                    batchsize=args.batchsize,
                )
        except:
            exit_and_clean_up(temp_folder)

        # Calculate the number of deduplicated reads
        deduplicated_reads = sum([d["nreads"] for d in abund])

        # Name the output file based on the input file
        # Ultimately adding ".json.gz" to the input file name
        if args.sample_name is not None:
            output_prefix = args.sample_name
        else:
            output_prefix = input_str[0].split("/")[-1]
        logging.info("Using sample name {} for output prefix".format(
            output_prefix))

        # Count the total number of reads
        logging.info("Counting the total number of reads")
        n_reads = count_fastq_reads(read_fp)
        logging.info("Reads in input file: {:,}".format(n_reads))

        # Read in the logs
        logging.info("Reading in the logs")
        logs = open(log_fp, 'rt').readlines()

        # Wrap up all of the results into a single JSON
        # and write it to the output folder
        output = {
            "input_path": "+".join(input_str),
            "input": output_prefix,
            "sample": args.sample_name,
            "output_folder": args.output_folder,
            "logs": logs,
            "ref_db": db_fp,
            "ref_db_url": args.ref_db,
            "results": abund,
            "total_reads": n_reads,
            "aligned_reads": aligned_reads,
            "deduplicated_reads": deduplicated_reads,
            "time_elapsed": time.time() - start_time,
            "params": {
                "batchsize": args.batchsize,
                "min_score": args.min_score,
                "blocks": args.blocks,
                "query_gencode": args.query_gencode,
                "threads": args.threads,
                "min_qual": args.min_qual
            }
        }
        return_results(
            output, output_prefix, args.output_folder, temp_folder
        )

        # Delete any files that were created for this sample
        logging.info("Removing temporary folder: " + temp_folder)
        shutil.rmtree(temp_folder)

        # Stop logging
        logging.info("Done")
        logging.shutdown()

    def filter(self):
        """Filter a set of alignments with FAMLI."""
        parser = argparse.ArgumentParser(
                description="""Filter a set of existing alignments in tabular
                format with FAMLI""")

        parser.add_argument("--input",
                            type=str,
                            help="Location for input alignement file.")
        parser.add_argument("--output",
                            type=str,
                            help="Location for output JSON file.")
        parser.add_argument("--threads",
                            type=int,
                            help="""Number of processors to use.""",
                            default=4)
        parser.add_argument("--logfile",
                            type=str,
                            help="""(Optional) Write log to this file.""")
        parser.add_argument("--batchsize",
                            type=int,
                            help="""Number of reads to process at a time.""")
        parser.add_argument("--qseqid-ix",
                            default=0,
                            type=int,
                            help="""Alignment column for query sequence ID.
                            (0-indexed column ix)""")
        parser.add_argument("--sseqid-ix",
                            default=1,
                            type=int,
                            help="""Alignment column for subject sequence ID.
                            (0-indexed column ix)""")
        parser.add_argument("--sstart-ix",
                            default=8,
                            type=int,
                            help="""Alignment column for subject start position.
                            (0-indexed column ix, 1-indexed start position)""")
        parser.add_argument("--send-ix",
                            default=9,
                            type=int,
                            help="""Alignment column for subject end position.
                            (0-indexed column ix, 1-indexed end position)""")
        parser.add_argument("--bitscore-ix",
                            default=11,
                            type=int,
                            help="""Alignment column for alignment bitscore.
                            (0-indexed column ix)""")
        parser.add_argument("--slen-ix",
                            default=13,
                            type=int,
                            help="""Alignment column for subject length.
                            (0-indexed column ix)""")
        parser.add_argument("--sd-mean-cutoff",
                            default=1.0,
                            type=float,
                            help="""Threshold for filtering max SD / MEAN""")
        parser.add_argument("--strim-5",
                            default=18,
                            type=int,
                            help="""Amount to trim from 5' end of subject""")
        parser.add_argument("--strim-3",
                            default=18,
                            type=int,
                            help="""Amount to trim from 3' end of subject""")

        args = parser.parse_args(sys.argv[2:])

        start_time = time.time()

        assert os.path.exists(args.input)

        # Set up logging
        logFormatter = logging.Formatter(
            '%(asctime)s %(levelname)-8s [FAMLI parse] %(message)s'
        )
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.INFO)

        if args.logfile:
            # Write to file
            fileHandler = logging.FileHandler(args.logfile)
            fileHandler.setFormatter(logFormatter)
            rootLogger.addHandler(fileHandler)

        # Write to STDOUT
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(logFormatter)
        rootLogger.addHandler(consoleHandler)

        if args.input.endswith(".gz"):
            f = gzip.open(args.input, "rt")
        else:
            f = open(args.input, "rt")

        aligned_reads, output = parse_alignment(
            f,
            QSEQID_i=args.qseqid_ix,
            SSEQID_i=args.sseqid_ix,
            SSTART_i=args.sstart_ix,
            SEND_i=args.send_ix,
            BITSCORE_i=args.bitscore_ix,
            SLEN_i=args.slen_ix,
            SD_MEAN_CUTOFF=args.sd_mean_cutoff,
            STRIM_5=args.strim_5,
            STRIM_3=args.strim_3,
            threads=args.threads,
            batchsize=args.batchsize,
        )

        f.close()

        if args.output:
            with open(args.output, "wt") as fo:
                json.dump(output, fo, indent=4)

        elapsed = round(time.time() - start_time, 2)
        logging.info("Time elapsed: {:,}".format(elapsed))


def main():
    """Entrypoint for main script."""
    FAMLI()


if __name__ == "__main__":
    # Run FAMLI
    FAMLI()



#!/usr/bin/python

import os
import gzip
import json
import time
import logging
import argparse
import numpy as np
from multiprocessing import Pool
from collections import defaultdict


class BLAST6Parser:
    """Object to help with parsing alignments in BLAST 6 format."""

    def __init__(self):
        # Keep track of the length of all subjects
        self.subject_len = {}
        # Keep track of the number of queries
        self.unique_queries = set([])

    def parse(self,
              align_handle,
              QSEQID_i=0,
              SSEQID_i=1,
              SSTART_i=8,
              SEND_i=9,
              EVALUE_i=10,
              BITSCORE_i=11,
              SLEN_i=13):
        """Parse a file, while keeping track of the subject lengths."""
        logging.info("Reading in alignments")

        for i, line in enumerate(align_handle):
            if i % 1000000 == 0 and i > 0:
                logging.info("{:,} lines of alignment parsed".format(i))
            line_list = line.rstrip("\n").split("\t")
            # Get the query and subject
            query = line_list[QSEQID_i]
            subject = line_list[SSEQID_i]
            bitscore = float(line_list[BITSCORE_i])
            sstart = int(line_list[SSTART_i])
            send = int(line_list[SEND_i])

            # Save information for the subject length
            if subject not in self.subject_len:
                self.subject_len[subject] = int(line_list[SLEN_i])

            # Add the query to the set of unique queries
            self.unique_queries.add(query)

            # Yield a tuple with query, subject, sstart, send, bitscore
            yield (
                query,
                subject,
                sstart - 1,  # Convert to 0-based indexing
                send,
                bitscore
            )

        if len(self.unique_queries) == 0:
            logging.info("Zero alignments were found")
        else:
            logging.info("Done reading in {:,} alignments".format(i + 1))
            logging.info("Number of unique subjects: {:,}".format(
                len(self.subject_len)))
            logging.info("Number of unique queries: {:,}".format(
                len(self.unique_queries)))

    def yield_alignments(
        self,
        align_handle,
        batchsize=None,
        QSEQID_i=0,
        SSEQID_i=1,
        SSTART_i=8,
        SEND_i=9,
        EVALUE_i=10,
        BITSCORE_i=11,
        SLEN_i=13
    ):
        """Yield batches of alignments, never to splitting up queries."""

        # If the batchsize function has been disabled, return the full list
        if batchsize is None:
            logging.info("Batchsize feature disabled, reading all alignments")
            yield [
                a
                for a in self.parse(
                    align_handle,
                    QSEQID_i=QSEQID_i,
                    SSEQID_i=SSEQID_i,
                    SSTART_i=SSTART_i,
                    SEND_i=SEND_i,
                    EVALUE_i=EVALUE_i,
                    BITSCORE_i=BITSCORE_i,
                    SLEN_i=SLEN_i
                )
            ]
            return

        logging.info("Reading in batches of {:,} alignments".format(batchsize))

        assert batchsize > 0, "Batches must be larger than 0"

        # Allocate a list that is 10% larger than the batch size
        alignments = [None] * int(batchsize * 1.1)
        # Keep track of how many alignments have been added to the list
        n_alignments_filled = 0
        # Keep track of the number of unique queries in this batch
        n_unique_queries = 0
        # Keep track of which batch of alignments we are on
        batch_num = 1
        # Keep track of the last unique query ID
        last_query = None

        # Add to the alignments, stopping when we get to the batch size limit
        for a in self.parse(
            align_handle,
            QSEQID_i=QSEQID_i,
            SSEQID_i=SSEQID_i,
            SSTART_i=SSTART_i,
            SEND_i=SEND_i,
            EVALUE_i=EVALUE_i,
            BITSCORE_i=BITSCORE_i,
            SLEN_i=SLEN_i
        ):

            # When we reach the batch size and new query, yield the alignments
            if a[0] != last_query and n_alignments_filled >= batchsize:
                logging.info("Processing batch {:,}".format(batch_num))
                logging.info("Alignments: {:,}".format(n_alignments_filled))
                logging.info("Unique queries: {:,}".format(n_unique_queries))

                if n_alignments_filled < int(batchsize * 1.1):
                    yield alignments[:n_alignments_filled]
                else:
                    yield alignments
                # Reset the alignment data
                alignments = [None] * int(batchsize * 1.1)
                n_alignments_filled = 0
                n_unique_queries = 0
                batch_num += 1

            # Add to the alignment list
            if n_alignments_filled < int(batchsize * 1.1):
                alignments[n_alignments_filled] = a
            else:
                alignments.append(a)

            # Increment the counter of filled alignments
            n_alignments_filled += 1
            # Increment the counter of unique queries
            if a[0] != last_query:
                n_unique_queries += 1
            # Record which query we saw in this alignment
            last_query = a[0]

        # Yield the final block of alignments
        logging.info("Processing batch {:,}".format(batch_num))
        logging.info("Alignments: {:,}".format(n_alignments_filled))
        logging.info("Unique queries {:,}".format(n_unique_queries))
        if n_alignments_filled < int(batchsize * 1.1):
            yield alignments[:n_alignments_filled]
        else:
            yield alignments


def recalc_subject_weight_worker(args):
    """Calculate the new weight of a subject."""
    query_weights, subject_len, subject_name = args
    return subject_name, sum(query_weights) / subject_len


class FAMLI_Reassignment:
    """Object to help with parsing alignments in BLAST 6 format."""

    def __init__(self, alignments, subject_len, pool=None):
        # Save the length of all subjects
        self.subject_len = subject_len

        # Keep track of the bitscores for each query/subject
        self.bitscores = defaultdict(lambda: defaultdict(float))

        # Keep track of the number of uniquely aligned queries
        self.n_unique = 0

        # Pool of workers
        if pool is None:
            self.pool = Pool(4)
        else:
            self.pool = pool

        logging.info("Adding alignments to read re-assignment model")
        for query, subject, sstart, send, bitscore in alignments:
            # Save information for the query and subject
            self.bitscores[query][subject] = bitscore

    def init_subject_weight(self):
        """Initialize the subject weights, equal and normalized to their length."""
        logging.info("Initializing subject weights")
        self.subject_weight = {
            subject: 1.0 / length
            for subject, length in self.subject_len.items()
        }

        # Also initialize the alignment probabilities as the bitscores
        logging.info("Initializing alignment probabilities")
        self.aln_prob = defaultdict(dict)
        self.aln_prob_T = defaultdict(dict)

        # Keep track of the queries and subjects that will need to be updated
        self.subjects_to_update = set([])
        self.queries_to_update = set(self.bitscores.keys())

        # Keep track of the queryies that have multiple alignments
        self.multimapped_queries = set(self.bitscores.keys())

        for query, bitscores in self.bitscores.items():
            if len(bitscores) == 1:
                self.n_unique += 1
            bitscore_sum = sum(bitscores.values())
            for subject, bitscore in bitscores.items():
                self.subjects_to_update.add(subject)
                v = bitscore / bitscore_sum
                self.aln_prob[query][subject] = v
                self.aln_prob_T[subject][query] = v

    def recalc_subject_weight(self):
        """Recalculate the subject weights."""
        self.queries_to_update = set([])

        self.queries_to_update = set([
            query
            for subject in self.subjects_to_update
            for query in self.aln_prob_T[subject].keys()
        ])

        for subject, new_weight in self.pool.imap(
            recalc_subject_weight_worker,
            [
                [
                    list(self.aln_prob_T[subject].values()),
                    self.subject_len[subject],
                    subject
                ]
                for subject in self.subjects_to_update
            ]
        ):
            self.subject_weight[subject] = new_weight

    def recalc_aln_prob(self):
        """Recalculate the alignment probabilities."""
        self.subjects_to_update = set([])

        # Iterate over every query
        for query in self.queries_to_update:
            aln_prob = self.aln_prob[query]
            new_probs = [
                prob * self.subject_weight[subject]
                for subject, prob in aln_prob.items()
            ]
            new_probs_sum = sum(new_probs)
            for ix, subject in enumerate(aln_prob):
                self.subjects_to_update.add(subject)
                v = new_probs[ix] / new_probs_sum
                self.aln_prob[query][subject] = v
                self.aln_prob_T[subject][query] = v

    def trim_least_likely(self, scale=0.9):
        """Remove the least likely alignments."""
        logging.info("Removing alignments below {} of the max".format(scale))
        n_trimmed = 0
        newly_unique_queries = set([])
        for query in self.multimapped_queries:
            aln_prob = self.aln_prob[query]
            # Skip queries with no remaining alignments
            if len(aln_prob) == 0:
                continue
            # Skip queries with only a single possible subject
            if len(aln_prob) == 1:
                newly_unique_queries.add(query)
                self.n_unique += 1
                continue
            # Figure out our maximum score for this query
            max_likely = max(list(aln_prob.values()))

            # Trim anyone BELOW the maximum possible value for this query.
            to_remove = [
                subject for subject, prob in aln_prob.items()
                if prob < scale * max_likely
            ]

            n_trimmed += len(to_remove)

            # Remove the subjects
            for subject in to_remove:
                del self.aln_prob[query][subject]
                del self.aln_prob_T[subject][query]

            if len(self.aln_prob[query]) == 1:
                newly_unique_queries.add(query)
                self.n_unique += 1

        self.multimapped_queries = self.multimapped_queries - newly_unique_queries

        logging.info("Removed {:,} unlikely alignments".format(n_trimmed))
        logging.info("Number of uniquely aligned queries: {:,}".format(
            self.n_unique))
        return n_trimmed


def filter_subjects_by_coverage(args):
    """Check whether the subject passes the coverage filter."""

    subject, cov, SD_MEAN_CUTOFF, STRIM_5, STRIM_3 = args

    # Trim the ends
    if cov.shape[0] >= STRIM_5 + STRIM_3 + 10:
        cov = cov[STRIM_5: -STRIM_3]

    passes_filter = cov.mean() > 0 and cov.std() / cov.mean() <= SD_MEAN_CUTOFF
    return subject, passes_filter


def calc_cov_by_subject(alignments, subject_len):
    """Index a set of sorted alignments by subject."""
    # The output will be a dict with the start:stop index for each subject
    assert len(alignments) > 0
    index = {}

    coverages = {
        subject: np.zeros(length, dtype=int)
        for subject, length in subject_len.items()
    }

    last_subject = None
    last_start_ix = None
    for ix, a in enumerate(alignments):
        # query, subject, sstart, send, bitstore = a
        if a[1] != last_subject:
            if last_subject is not None:
                index[last_subject] = (last_start_ix, ix)
            last_subject = a[1]
            last_start_ix = ix

        # Add to the cov_proc
        coverages[a[1]][a[2]:a[3]] += 1

    index[last_subject] = (last_start_ix, ix)

    return coverages, index


def parse_alignment(align_handle,
                    batchsize=None,   # Number of reads to process at a time
                    QSEQID_i=0,
                    SSEQID_i=1,
                    SSTART_i=8,
                    SEND_i=9,
                    BITSCORE_i=11,
                    SLEN_i=13,
                    SD_MEAN_CUTOFF=1.0,
                    STRIM_5=18,
                    STRIM_3=18,
                    threads=4,
                    MAX_ITERATIONS=1000):
    """
    Parse an alignment in BLAST6 format and determine which subjects are likely to be present. This is the core of FAMLI.
    BLAST 6 columns by default (in order): qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen
                                              0     1       2       3       4       5       6   7   8       9   10      11      12   13  
    """
    # Initialize the multithreading pool before reading in the alignments
    pool = Pool(threads)

    # Initialize the alignment parser
    parser = BLAST6Parser()

    # Keep a list of the final, filtered alignments
    final_alignments = []

    # Iterate over blocks of alignments
    for alignments in parser.yield_alignments(
        align_handle,
        batchsize=batchsize,
        QSEQID_i=QSEQID_i,
        SSEQID_i=SSEQID_i,
        SSTART_i=SSTART_i,
        SEND_i=SEND_i,
        BITSCORE_i=BITSCORE_i,
        SLEN_i=SLEN_i
    ):
        if len(alignments) == 0:
            logging.info("Zero alignments to process in this batch")
            continue

        # Count the total number of reads that were aligned
        logging.info("Counting unique queries")
        aligned_reads = len(set([a[0] for a in alignments]))

        logging.info("Number of unique queries: {:,}".format(aligned_reads))
        logging.info("Number of alignments: {:,}".format(len(alignments)))

        # Sort alignments by subject
        logging.info("Sorting alignments by subject")
        alignments.sort(key=lambda a: a[1])

        logging.info("Calculating coverage by subject")
        subject_coverages, subject_index = calc_cov_by_subject(
            alignments, parser.subject_len)

        # STEP 1. FILTER SUBJECTS BY "COULD" COVERAGE EVENNESS
        logging.info("FILTER 1: Even coverage of all alignments")

        filter_1 = pool.map(filter_subjects_by_coverage, [
            [
                subject,
                coverage,
                SD_MEAN_CUTOFF,
                STRIM_5,
                STRIM_3
            ]
            for subject, coverage in subject_coverages.items()
        ])

        # Reformat as a dict
        filter_1 = dict(filter_1)

        logging.info("Subjects passing FILTER 1: {:,} / {:,}".format(
            sum(filter_1.values()), len(filter_1)
        ))

        # Subset the alignments to only those subjects passing the filter
        alignments = [
            a
            for subject, start_stop in subject_index.items()
            for a in alignments[start_stop[0]: start_stop[1]]
            if filter_1[subject]
        ]

        logging.info("Queries passing FILTER 1: {:,} / {:,}".format(
            len(set([a[0] for a in alignments])), aligned_reads
        ))

        # STEP 2: Reassign multi-mapped reads to a single subject
        logging.info("FILTER 2: Reassign queries to a single subject")

        # Add the alignments to a model to optimally re-assign reads
        model = FAMLI_Reassignment(alignments, parser.subject_len, pool=pool)

        # Initialize the subject weights
        model.init_subject_weight()

        ix = 0
        while ix <= MAX_ITERATIONS:
            ix += 1
            logging.info("Iteration: {:,}".format(ix))
            # Recalculate the subject weight, given the naive alignment probabliities
            model.recalc_subject_weight()
            # Recalculate the alignment probabilities, given the subject weights
            model.recalc_aln_prob()

            # Trim the least likely alignment for each read
            n_trimmed = model.trim_least_likely()

            if n_trimmed == 0:
                break

        logging.info("Subsetting the alignment to only the reassigned queries")

        # Subset the alignment to only the reassigned queries
        alignments = [
            a
            for a in alignments
            if a[1] in model.aln_prob[a[0]] and
            len(model.aln_prob[a[0]]) == 1
        ]

        logging.info("Finished reassigning reads ({:,} remaining)".format(
            len(alignments)))

        # Add to the batch of final alignments
        final_alignments.extend(alignments)
        logging.info("A total of {:,} filtered alignments collected".format(
            len(final_alignments)))
        del alignments

    if len(final_alignments) == 0:
        logging.info("The entire sample contains zero alignments")
        # Return some empty data
        return 0, []

    # Change the name, for brevity
    alignments = final_alignments
    queries_after_reassignment = len(alignments)

    # Cull the subjects that were removed during FILTER 2
    all_subjects = set([a[1] for a in alignments])
    parser.subject_len = {
        subject: length
        for subject, length in parser.subject_len.items()
        if subject in all_subjects
    }

    # STEP 3: Filter subjects by even coverage of reassigned queries
    logging.info("FILTER 3: Filtering subjects by even sequencing coverage")

    subject_coverages, subject_index = calc_cov_by_subject(
        alignments, parser.subject_len)

    filter_3 = pool.map(filter_subjects_by_coverage, [
        [
            subject,
            coverage,
            SD_MEAN_CUTOFF,
            STRIM_5,
            STRIM_3
        ]
        for subject, coverage in subject_coverages.items()
    ])
    # Reformat as a dict
    filter_3 = dict(filter_3)

    logging.info("Subjects passing FILTER 3: {:,} / {:,}".format(
        sum(filter_3.values()), len(filter_3)
    ))

    # Subset the alignments to only those subjects passing the filter
    alignments = [
        a
        for subject, start_stop in subject_index.items()
        for a in alignments[start_stop[0]: start_stop[1]]
        if filter_3[subject]
    ]

    logging.info("Queries passing FILTER 3: {:,} / {:,}".format(
        len(set([a[0] for a in alignments])), queries_after_reassignment
    ))

    # Make the output by calculating coverage per subject
    output = []

    # Make a dict of the alignment ranges
    logging.info("Collecting final alignments")
    alignment_ranges = defaultdict(list)
    for query, subject, sstart, send, bitscore in alignments:
        alignment_ranges[subject].append((sstart, send))

    logging.info("Calculating final stats")

    for subject, aln_ranges in alignment_ranges.items():
        # Make a coverage map
        cov = np.zeros(model.subject_len[subject], dtype=int)

        # Add to the coverage
        for sstart, send in aln_ranges:
            cov[sstart:send] += 1

        output.append({
            "id": subject,
            "nreads": len(aln_ranges),
            "coverage": (cov > 0).mean(),
            "depth": cov.mean(),
            "std": cov.std(),
            "length": model.subject_len[subject],
        })

    logging.info("Results: assigned {:,} queries to {:,} subjects".format(
        sum([d["nreads"] for d in output]),
        len(output),
    ))

    return aligned_reads, output



"""Setup module for FAMLI"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='famli',
    version='1.0',
    description='Functional Analysis of Metagenomes by Likelihood Inferrence',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/FredHutch/FAMLI',
    author='Samuel Minot',
    author_email='sminot@fredhutch.org',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='metagenomics microbiome science',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        "pandas>=0.20.3",
        "biopython>=1.70",
        "numpy>=1.13.1",
        "scipy>=0.19.1",
        "awscli>=1.11.146",
        "boto3>=1.4.7",
        "python-dateutil==2.6.0"
    ],
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/fredhutch/famli/issues',
        'Source': 'https://github.com/fredhutch/famli/',
    },
    entry_points={
        'console_scripts': [
            'famli = famli.run_famli:main',
        ],
    },
)